/* Routines for reading in line6 .lib files */

#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include "libfile.h"

static void delay_time_to_controls(double delay_time, int *course, int *fine)
{
    double c_scale = 3150.0 / 128.0;
    double f_scale = 3150.0 / (128.0 * 128.0);
    double rem;

    *course = (int) (delay_time / c_scale);
    rem = delay_time - (*course * c_scale);

    *fine = (int) (rem / f_scale);
}

/* convert time from packed format to a time_t */
static time_t convert_time(unsigned int packed_time)
{
    struct tm time_val;

    time_val.tm_sec  = (packed_time & 0x1f) * 2;
    time_val.tm_min  = ((packed_time >>  5) & 0x3f);
    time_val.tm_hour = ((packed_time >> 11) & 0x1f);
    time_val.tm_mday = ((packed_time >> 16) & 0x1f);
    time_val.tm_mon  = ((packed_time >> 21) & 0x0f) - 1;
    time_val.tm_year = ((packed_time >> 25) & 0x7f) + 80;

    return mktime(&time_val);
}

/* read an int with some endianness and size into host byte order */
static int get_host_int(unsigned char *buffer, int num_bytes, 
                        gboolean is_bigendian)
{
    int i;
    int host_int = 0;
    int shift;
    int shift_inc;

    g_assert (num_bytes <= sizeof(int));

    if (is_bigendian)
    {
        shift = 8 * (num_bytes-1);
        shift_inc = -8;
    }
    else
    {
        shift = 0;
        shift_inc = 8;
    }

    for (i=0; i<num_bytes; i++) 
    {
        host_int |= buffer[i] << shift;
        shift += shift_inc;
    }

    /* sign extension if num_bytes < sizeof(int) */
    if (host_int & (1 << (num_bytes * 8 - 1)))
    {
        host_int |= (-1 << (num_bytes * 8 - 1));
    }
    return host_int;
}

static gboolean check_file(libfile_t *file)
{
    /* whoever made this file bisexual in byte order should die...
       even the strings are backwards! */

    if ((strcmp(file->form.form, "FORM") != 0) &&
        (strcmp(file->form.form, "MROF") != 0))
        return FALSE;

    if ((strcmp(file->form.sslb, "SSLB") != 0) &&
        (strcmp(file->form.sslb, "BLSS") != 0))
        return FALSE;

    if ((strcmp(file->lhdr.lhdr, "LHDR") != 0) &&
        (strcmp(file->lhdr.lhdr, "RDHL") != 0) &&
        (strlen(file->lhdr.lhdr) > 0))
        return FALSE;

    if ((strcmp(file->wseq.wseq, "WSEQ") != 0) &&
        (strcmp(file->wseq.wseq, "QESW") != 0) &&
        (strlen(file->wseq.wseq) > 0))
        return FALSE;

    if ((strcmp(file->lent.lent, "LENT") != 0) &&
        (strcmp(file->lent.lent, "TNEL") != 0) &&
        (strlen(file->lent.lent) > 0))
        return FALSE;

    return TRUE;
}

static gboolean read_form(FILE *fp, libfile_t *file)
{
    char buf[4];

    if (fread(file->form.form, 4, 1, fp) != 1)
        return FALSE;

    file->is_big_endian = (strcmp(file->form.form, "FORM") == 0);

    if (fread(buf, 4, 1, fp) != 1)
        return FALSE;

    file->form.size = get_host_int(buf, 4, file->is_big_endian);

    if (fread(file->form.sslb, 4, 1, fp) != 1)
        return FALSE;

    return TRUE;
}

static gboolean read_lhdr(FILE *fp, libfile_t *file)
{
    char buf[4];

    if (fread(buf, 4, 1, fp) != 1)
        return FALSE;

    file->lhdr.size = get_host_int(buf, 4, file->is_big_endian);

    file->lhdr.data = g_malloc(file->lhdr.size);

    if (fread(file->lhdr.data, file->lhdr.size, 1, fp) != 1)
        return FALSE;

    return TRUE;
}

static gboolean read_wseq(FILE *fp, libfile_t *file)
{
    char buf[4];

    if (fread(buf, 4, 1, fp) != 1)
        return FALSE;

    file->wseq.size = get_host_int(buf, 4, file->is_big_endian);

    file->wseq.data = g_malloc(file->wseq.size);

    if (fread(file->wseq.data, file->wseq.size, 1, fp) != 1)
        return FALSE;

    return TRUE;
}


static gboolean read_lent(FILE *fp, libfile_t *file)
{
    guchar buf[4];
    int i;
    char ch;
    int start_pos, remaining;
    int pack_date;
    guint unscaled_val;
    double delay_time; 
    int d_coarse, d_fine;

    if (fread(buf, 4, 1, fp) != 1)
        return FALSE;

    file->lent.size = get_host_int(buf, 4, file->is_big_endian);

    start_pos = ftell(fp);
    if (fread(file->lent.pad1, 4, 1, fp) != 1)
        return FALSE;

    if (fread(buf, 4, 1, fp) != 1)
        return FALSE;

    pack_date = get_host_int(buf, 4, file->is_big_endian);
    file->lent.date = convert_time(pack_date);

    if (fread(file->lent.pad2, 5, 1, fp) != 1)
        return FALSE;

    file->lent.title_size = fgetc(fp);

    if (fread(file->lent.title, file->lent.title_size, 1, fp) != 1)
        return FALSE;

    if (fread(file->lent.pad2, 2, 1, fp) != 1)
        return FALSE;

    file->lent.params.distortion_on = fgetc(fp);
    file->lent.params.drive_boost_on = fgetc(fp);
    file->lent.params.presence_on = fgetc(fp);
    file->lent.params.delay_on = fgetc(fp);
    file->lent.params.rotary_on = fgetc(fp);
    file->lent.params.reverb_on = fgetc(fp);
    file->lent.params.noise_gate_on = fgetc(fp);
    file->lent.params.bright_on = fgetc(fp);

    file->lent.params.amp_type = fgetc(fp);

    file->lent.params.amp_type = MIN(file->lent.params.amp_type, 32);

    file->lent.params.drive = fgetc(fp);
    file->lent.params.drive2 = fgetc(fp);

    file->lent.params.bass = fgetc(fp);
    file->lent.params.mid = fgetc(fp);
    file->lent.params.treble = fgetc(fp);
    file->lent.params.presence = fgetc(fp);
    file->lent.params.channel_volume = fgetc(fp);
    file->lent.params.gate_threshold = fgetc(fp);
    file->lent.params.gate_decay = fgetc(fp);
    file->lent.params.wah_current = fgetc(fp);
    file->lent.params.wah_bottom = fgetc(fp);
    file->lent.params.wah_top = fgetc(fp);

    fgetc(fp);

    file->lent.params.vol_current = fgetc(fp);
    file->lent.params.vol_minimum = fgetc(fp);
    file->lent.params.vol_type = fgetc(fp);

    if (fread(file->lent.pad5, 2, 1, fp) != 1)
        return FALSE;

    if (fread(buf, 3, 1, fp) != 1)
        return FALSE;

    unscaled_val = (buf[0] << 16) | (buf[1] << 8) | buf[2];
    delay_time = (double) unscaled_val * 3150.0 / 98298.0;

    file->lent.params.delay_unscaled = unscaled_val;
    file->lent.params.delay_time = (int) (delay_time + 0.5);
    delay_time_to_controls(delay_time, &d_coarse, &d_fine);
    file->lent.params.delay_coarse = (guint8) d_coarse;
    file->lent.params.delay_fine = (guint8) d_fine;

    if (fread(file->lent.pad5, 4, 1, fp) != 1)
        return FALSE;

    file->lent.params.delay_feedback = fgetc(fp);
    fgetc(fp);
    file->lent.params.delay_level = fgetc(fp);

    fgetc(fp);
    ch = fgetc(fp);
    file->lent.params.reverb_type = MIN(ch, 1);

    file->lent.params.reverb_decay = fgetc(fp); 
    file->lent.params.reverb_tone = fgetc(fp); 
    file->lent.params.reverb_diffuse = fgetc(fp); 
    file->lent.params.reverb_density = fgetc(fp); 
    file->lent.params.reverb_level = fgetc(fp);

    file->lent.params.cab_type = fgetc(fp);
    file->lent.params.cab_type = MIN(file->lent.params.cab_type, 15);

    file->lent.params.air = fgetc(fp);
    file->lent.params.effect = fgetc(fp);
    file->lent.params.effect = MIN(file->lent.params.effect, 15);

    file->lent.params.rotary_depth = fgetc(fp);

    if (fread(buf, 5, 1, fp) != 1)
        return FALSE;

    memcpy (file->lent.params.effect_params, buf, 5);
    file->lent.params.compression_ratio = MIN(buf[0], 5);

    file->lent.params.rotary_speed = (buf[0] != 0);
    unscaled_val = (buf[1] << 8) | buf[2];
    file->lent.params.rotary_speed_max = (2894 - unscaled_val)/22;
    unscaled_val = (buf[3] << 8) | buf[4];
    file->lent.params.rotary_speed_min = (2894 - unscaled_val)/22;

    file->lent.params.chorus_speed = (int)
      ((double) buf[1] * 127.0 / 200.0);

    unscaled_val = ((buf[2] << 8) | (buf[3]));

    /* scale to 0-127 */ 
    file->lent.params.chorus_depth = (int) 
      ((double) unscaled_val * 127.0 / 313.0);

    file->lent.params.chorus_feedback = buf[4];

    file->lent.params.tremolo_speed = buf[0];
    file->lent.params.tremolo_depth = buf[1];

    file->lent.params.swell_amplitude = buf[0];

    if (fread(buf, 2, 1, fp) != 1)
        return FALSE;

    unscaled_val = ((buf[0] << 8) | (buf[1]));
    file->lent.params.chorus_predelay_unscaled = unscaled_val;
    file->lent.params.chorus_predelay = (int)
      ((double) unscaled_val * 127.0 / 597.0);


    fgetc(fp);
    file->lent.location_size = fgetc(fp);

    if (fread(file->lent.location, file->lent.location_size, 1, fp) != 1)
        return FALSE;

    /* comment */
    if (fgetc(fp) == 2)
    {
        file->lent.comment_size = fgetc(fp);
        if (fread(file->lent.comment, file->lent.comment_size, 1, fp) != 1)
            return FALSE;
    }

    /* drop the rest */
    remaining = ftell(fp) - start_pos;
    for (i=0; i<remaining; i++)
        fgetc(fp);

    return TRUE;
}

static gboolean process_blocks(FILE *fp, libfile_t *file)
{
    char block_id[5];

    gboolean (*process_fn)(FILE *, libfile_t *);

    /* Read four bytes, then pick a function to call... */
 
    block_id[4] = 0; 
    while (!feof(fp))
    { 
        if (fread(block_id, 4, 1, fp) != 1)
            break;

        if ((strcmp(block_id, "WSEQ") == 0) ||
            (strcmp(block_id, "QESW") == 0))
        {
            strcpy(file->wseq.wseq, block_id);
            process_fn = read_wseq;
        }
        else if ((strcmp(block_id, "LENT") == 0) ||
                 (strcmp(block_id, "TNEL") == 0))
        {
            strcpy(file->lent.lent, block_id);
            process_fn = read_lent;
        }
        else if ((strcmp(block_id, "LHDR") == 0) ||
                 (strcmp(block_id, "RDHL") == 0))
        {
            strcpy(file->lhdr.lhdr, block_id);
            process_fn = read_lhdr;
        }
        else
        {
            fprintf (stderr, "Unknown blocktype: %s\n", block_id);
            return FALSE;
        }

        if (!process_fn(fp, file))
            return FALSE;
    } 
    return TRUE;
}

libfile_t *libfile_new(char *filename)
{
    FILE *fp;
    libfile_t *file;

    fp = fopen(filename, "rb");
    if (!fp) return NULL;
    
    /* alloc0 takes care of most string termination */
    file = g_malloc0(sizeof(libfile_t));

    /* read blocks as we find them */

    if (!read_form(fp, file) ||
        !process_blocks(fp, file) ||
        !check_file(file))
    {
        return NULL;
        libfile_destroy(file);
        return NULL;
    }

    fclose(fp);
    return file;
}

void libfile_destroy(libfile_t *libfile)
{
    g_free(libfile);
}

void libfile_print(libfile_t *file)
{
    printf ("Form: %s\n", file->form.form);
    printf ("  size: %d\n", file->form.size);
    printf ("  sslb: %s\n", file->form.sslb);

    printf ("\n");

    printf ("  lhdr: %s\n", file->lhdr.lhdr);
    printf ("    size: %d\n", file->lhdr.size);
    printf ("\n");

    printf ("  wseq: %s\n", file->wseq.wseq);
    printf ("    size: %d\n", file->wseq.size);
    printf ("\n");

    printf ("  lent: %s\n", file->lent.lent);
    printf ("    size: %d\n", file->lent.size);
    printf ("    date: %s", ctime(&file->lent.date));
    printf ("    title_size: %d\n", file->lent.title_size);
    printf ("    title: %s\n", file->lent.title);
    printf ("    distortion: %s\n", file->lent.params.distortion_on ? "on" : "off");
    printf ("    drive_boost: %s\n", file->lent.params.drive_boost_on ? "on" : "off");
    printf ("    presence: %s\n", file->lent.params.presence_on ? "on" : "off");
    printf ("    delay: %s\n", file->lent.params.delay_on ? "on" : "off");
    printf ("    rotary: %s\n", file->lent.params.rotary_on ? "on" : "off");
    printf ("    reverb: %s\n", file->lent.params.reverb_on ? "on" : "off");
    printf ("    noise_gate: %s\n", file->lent.params.noise_gate_on ? "on" : "off");
    printf ("    bright: %s\n", file->lent.params.bright_on ? "on" : "off");
    printf ("    amp: %s\n", _amp_model_table[file->lent.params.amp_type]);
    printf ("    cab: %s\n", _cab_model_table[file->lent.params.cab_type]);
    printf ("    air: %d\n", file->lent.params.air);
    printf ("    drive: %d\n", file->lent.params.drive);
    printf ("    drive2: %d\n", file->lent.params.drive2);
    printf ("    bass: %d\n", file->lent.params.bass);
    printf ("    mid: %d\n", file->lent.params.mid);
    printf ("    treble: %d\n", file->lent.params.treble);
    printf ("    presence: %d\n", file->lent.params.presence);
    printf ("    channel_volume: %d\n", file->lent.params.channel_volume);
    printf ("    gate_threshold: %d\n", file->lent.params.gate_threshold);
    printf ("    gate_decay: %d\n", file->lent.params.gate_decay);
    printf ("    wah_current: %d\n", file->lent.params.wah_current);
    printf ("    wah_bottom: %d\n", file->lent.params.wah_bottom);
    printf ("    wah_top: %d\n", file->lent.params.wah_top);
    printf ("    vol_current: %d\n", file->lent.params.vol_current);
    printf ("    vol_minimum: %d\n", file->lent.params.vol_minimum);
    printf ("    vol_type: %s\n", _vol_type_table[file->lent.params.vol_type]);
    printf ("    reverb_type: %s\n", _reverb_type_table[file->lent.params.reverb_type]);
    printf ("    reverb_decay: %d\n", file->lent.params.reverb_decay);
    printf ("    reverb_tone: %d\n", file->lent.params.reverb_tone);
    printf ("    reverb_diffuse: %d\n", file->lent.params.reverb_diffuse);
    printf ("    reverb_density: %d\n", file->lent.params.reverb_density);
    printf ("    reverb_level: %d\n", file->lent.params.reverb_level);
    printf ("    delay_time: %d\n", file->lent.params.delay_time);
    printf ("    delay_coarse: %d\n", file->lent.params.delay_coarse);
    printf ("    delay_fine: %d\n", file->lent.params.delay_fine);
    printf ("    delay_feedback: %d\n", file->lent.params.delay_feedback);
    printf ("    delay_level: %d\n", file->lent.params.delay_level);
    printf ("    effect: %s\n", _effect_name_table[_effect_translation_table[file->lent.params.effect]]);
    printf ("    compression_ratio: %s\n", _compression_ratio_table[file->lent.params.compression_ratio]);
    printf ("    tremolo_speed: %d\n", file->lent.params.tremolo_speed);
    printf ("    tremolo_depth: %d\n", file->lent.params.tremolo_depth);
    printf ("    chorus_predelay: %d\n", file->lent.params.chorus_predelay);
    printf ("    chorus_speed: %d\n", file->lent.params.chorus_speed);
    printf ("    chorus_depth: %d\n", file->lent.params.chorus_depth);
    printf ("    chorus_feedback: %d\n", file->lent.params.chorus_feedback);
    printf ("    rotary_depth: %d\n", file->lent.params.rotary_depth);
    printf ("    rotary_speed: %s\n", (file->lent.params.rotary_speed) ?
                                      "Fast" : "Slow");
    printf ("    rotary_min: %d\n", file->lent.params.rotary_speed_min);
    printf ("    rotary_max: %d\n", file->lent.params.rotary_speed_max);
    printf ("    swell_amplitude: %d\n", file->lent.params.swell_amplitude);
    printf ("    location: %s\n", file->lent.location);

    if (file->lent.comment_size > 0)
    {
        printf ("    comment: %s\n", file->lent.comment);
    }
}
