/* routines to send midi commands for controls */

#include "midi_xfer.h"
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>

/* default device */
#define MIDI_DEVICE "/dev/midi"

/* expand 8 bit values to two bytes */
#define exp_byte(s, c)  \
{  \
    *(s)++ = ((c) >> 4) & 0x0f;  \
    *(s)++ = (c) & 0x0f;  \
}

static char control_codes[] =
{
    12, 13, 20, 14, 15, 16, 73, 21, 17, 22, 23, 24, 19, 
    1, 25, 26, 27, 28, 30, 62, 32, 34, 36, 37, 38, 39, 
    40, 41, 18, 42, 43, 4, 44, 45, 7, 46, 47, 48, 49, 
    64, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 71, 72
};

static void set_control(midi_context_t *m, char control, char value)
{
    char cmd[3];

    cmd[0] = 0xb0;
    cmd[1] = control;
    cmd[2] = value;

    write(m->fd, cmd, 3);
}

char *midi_find_device()
{
    DIR *dir;
    struct dirent *entry;
    char *filename;
    gboolean found_dev = FALSE;
    int fd;

    if (filename = getenv("MIDI_DEVICE"))
        return (g_strdup(filename));

    if (open(MIDI_DEVICE, O_WRONLY) != -1)
        return (g_strdup(MIDI_DEVICE));

    /* search through /dev for something that might work */
    if (dir = opendir("/dev"))
    {
        while ((entry = readdir(dir)) && !found_dev)
        {
            if (strncmp(entry->d_name, "midi", 4) == 0)
            {
                filename = g_strdup_printf("/dev/%s", entry->d_name);

                if ((fd = open(filename, O_WRONLY)) != -1)
                {
                    found_dev = TRUE;
                    close(fd);
                }
                else 
                    g_free(filename);
            }
        }
        closedir(dir);
    }

    if (found_dev) 
        return filename;

    return NULL;
}

midi_context_t *midi_open(char *device)
{
    midi_context_t *m = g_malloc(sizeof(midi_context_t));

    m->fd = open(device, O_WRONLY);
    if (m->fd == -1)
    {
        g_free(m);
        return NULL;
    }
    return m;
}

void midi_close(midi_context_t *m)
{
    close(m->fd);
    g_free(m);
}

/* This wrapper handles any scaling/translation we need to do */
void midi_set_param(midi_context_t *m, control_type_t control, int value)
{
    int scaled_value;
    int i;

    switch (control)
    {

        case CT_COMPRESSION:
            scaled_value = 23 * value;
            break;

        case CT_GATE_THRESH:
            /* for some reason this ranges 0-96 */
            scaled_value = (int) ((double) value * 127.0 / 96.0);
            break;

        case CT_EFFECT_SELECT:
        case CT_DRIVE:
        case CT_DRIVE_2:
        case CT_BASS:
        case CT_MID:
        case CT_TREBLE:
        case CT_PRESENCE:
        case CT_CHANNEL_VOL:
        case CT_GATE_DECAY:
        case CT_DELAY_REPEATS:
        case CT_DELAY_LEVEL:
        case CT_REVERB_DECAY:
        case CT_REVERB_TONE:
        case CT_REVERB_DIFFUSE:
        case CT_REVERB_DENSITY:
        case CT_REVERB_LEVEL:
        case CT_AIR:
            scaled_value = value * 2;
            break;

        case CT_SWELL:
        case CT_ROTARY_SPEED:
        case CT_BRIGHT:
        case CT_EQ:
        case CT_REVERB:
        case CT_VOLUME_LOCATION:
        case CT_DISTORTION:
        case CT_BOOST:
        case CT_DELAY:
        case CT_NOISE_GATE:
            scaled_value = value ? 127 : 0;
            break;
        default:
            scaled_value = value;
            break;
        
    }
    set_control(m, control_codes[control], scaled_value);
}

void midi_send_params(midi_context_t *m, amp_params_t *params)
{
    #define SYSEX_LEN 151

    char sysex_buf[SYSEX_LEN] = { 0xf0, 0x00, 0x01, 0x0c, 0x01, 0x01, 
                                  0x01, 0x00 };
    char *s;
    int delay, predelay, i;

    s = &sysex_buf[8];

    memset(s, 0, SYSEX_LEN-8);

    exp_byte(s, params->distortion_on);
    exp_byte(s, params->drive_boost_on);
    exp_byte(s, params->presence_on);
    exp_byte(s, params->delay_on);
    exp_byte(s, params->rotary_on);
    exp_byte(s, params->reverb_on);
    exp_byte(s, params->noise_gate_on);
    exp_byte(s, params->bright_on);

    exp_byte(s, params->amp_type);
    exp_byte(s, params->drive);
    exp_byte(s, params->drive2);
    exp_byte(s, params->bass);
    exp_byte(s, params->mid);
    exp_byte(s, params->treble);
    exp_byte(s, params->presence);
    exp_byte(s, params->channel_volume);

    exp_byte(s, params->gate_threshold);
    exp_byte(s, params->gate_decay);

    exp_byte(s, params->wah_current); 
    exp_byte(s, params->wah_bottom); 
    exp_byte(s, params->wah_top);
    
    exp_byte(s, 0); /* ?? */

    exp_byte(s, params->vol_current); 
    exp_byte(s, params->vol_minimum); 
    exp_byte(s, params->vol_type); 

    exp_byte(s, 0); /* ?? */

    delay = params->delay_unscaled;

    for (i=0; i<2; i++)
    {
        exp_byte(s, (delay >> 24));
        exp_byte(s, (delay >> 16)); 
        exp_byte(s, (delay >> 8));
        exp_byte(s, (delay));
    }
    exp_byte(s, params->delay_feedback);
    exp_byte(s, params->delay_feedback);
    exp_byte(s, params->delay_level);
    exp_byte(s, params->delay_level);

    exp_byte(s, params->reverb_type);
    exp_byte(s, params->reverb_decay);
    exp_byte(s, params->reverb_tone);
    exp_byte(s, params->reverb_diffuse);
    exp_byte(s, params->reverb_density);
    exp_byte(s, params->reverb_level);

    exp_byte(s, params->cab_type);
    exp_byte(s, params->air);

    /* FIXME: wrong */
    exp_byte(s, params->effect);

    exp_byte(s, params->rotary_depth);

    for (i=0; i<5; i++)
        exp_byte(s, params->effect_params[i]);

    predelay = params->chorus_predelay_unscaled;
    exp_byte(s, (predelay >> 8));
    exp_byte(s, (predelay));

    /* no idea on the rest of these bytes... */
     
    sysex_buf[SYSEX_LEN-1] = 0xf7;

    write(m->fd, sysex_buf, SYSEX_LEN);
}
